/******************************************************************************
bin.c, uses ems memory - 4 x 4 binning
Stores full 512X512 image in EMS
      Image Acquisition & Display Program - Primitive Control Functions
                       Runs on any VGA or MCGA display.
 Displays the image with the serial register on the right edge of the screen.
******************************************************************************/
#include <bios.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <at200.h>
#include <at200oem.h>
#include <camera.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#define DEFAULT_CCD      4            /* Thomson 512 x 512 CCD */
#define DEFAULT_ANALOG   A200KHZ      /* 200KHz 12 bit analog electronics */
#define DEFAULT_EXPOSURE 200          /* Exposure time in milliseconds */
#define DEFAULT_SHUTTER  31           /* Shutter open/close time in milliseconds */

int ccd_type = DEFAULT_CCD,
    ser_byte = DEFAULT_ANALOG,
    exposure = DEFAULT_EXPOSURE,
    shutter  = DEFAULT_SHUTTER,
    time = 1,
    loop = 1,
    ems_handle = 0;                   /* the ems handle */
    unsigned far *data;               /* pointer to ems data page */
int p_bin = 1;
int s_bin = 1;
char fname[20],name[20];                      /* Pointer to output file */
unsigned huge buffer[16*1024];
unsigned far *ptable;

char far *screen_row[200] = { 0, };   /* Pointers to graphics display rows */
/* screen_row[] is initialized in main() to point to the start of each row on
   the VGA display.  It is then used by the show_image() function to update
   the VGA memory. */
    int dofile = 0;

long dos_ticks(void)
{ /* Return the BIOS clock tick counter value */
  return( *((long far *) 0x46CL) );
}

void mcga_display_mode(int mode)
{ /* Select text mode or 320 x 200 - 256 color graphics mode */
union REGS regs;

  regs.x.ax = (mode) ? 19 : 3;  /* AH=0, AL=Mode */
  int86(16, &regs, &regs);      /* Call video bios */
}

void mcga_color_set(int color,int red,int green,int blue)
{ /* Set the RGB values for a color in 256 color graphics modes */
  outp(0x3C8,color);  /* DAC color control register selects the color */
  outp(0x3C9,red);    /* Set Red DAC (0..63) */
  outp(0x3C9,green);  /* Set Green DAC (0..63) */
  outp(0x3C9,blue);   /* Set Blue DAC (0..63) */
}

void load_ops(void)
{
unsigned x,rows,cols;

  rows = ccd[ccd_type].rows / p_bin;
  cols = ccd[ccd_type].cols / s_bin;
  disable_trigger();
  outp(at200_base + SER_CONTROL,ser_byte);
  x = at200_base + CONTROL;
  outp(x,inp(x) & ~FIFO_INPUT_SELECT);
  dsp_command( SET_MPP_EXPOSE );
  dsp_command( SET_MPP_CLOCK );
  dsp_command( SET_PARALLEL_DELAY );
  dsp_arg( (long) ccd[ccd_type].parallel_delay );
  dsp_command( SET_PARALLEL_STATES );
  dsp_arg( (long) ccd[ccd_type].parallel_states );
  dsp_command( SET_SHUTTER_CLOSE );
  dsp_arg( (long) shutter );
  dsp_command( SET_SHUTTER_OPEN );
  dsp_arg( (long) shutter );
  dsp_command( LOAD_OPERATIONS );
  dsp_op( OP_SET_GAIN,1 );
  dsp_op( OP_SET_SER_BIN,s_bin );
  dsp_op( OP_EXPOSE_SCALE,exposure ); /* Exposure Timer Scale in msec */
  dsp_op( OP_PARALLEL_DIS,ccd[ccd_type].rows );
  dsp_op( OP_DO_LOOP,loop );
    dsp_op( OP_CLEAR_UNTIL_TRIG,0 );
    dsp_op( OP_OPEN_SHUTTER,0 );
    dsp_op( OP_EXPOSE,time );            /* Expose time * Timer Scale */
    dsp_op( OP_CLOSE_SHUTTER,0 );
    x = ccd[ccd_type].cols + ccd[ccd_type].prescan + ccd[ccd_type].postscan;
    dsp_op( OP_SERIAL_DIS,x * 5);     /* Clear Serial Register */
    dsp_op( OP_DO_LOOP,512/p_bin );
      dsp_op( OP_PARALLEL_BIN,p_bin );
      dsp_op( OP_SERIAL_DIS,ccd[ccd_type].prescan);
      dsp_op( OP_SERIAL_CONV,512/s_bin );
      dsp_op( OP_SERIAL_DIS,ccd[ccd_type].postscan);
    dsp_op( OP_END_DO,0 );
  dsp_op( OP_END_DO,0 );
  dsp_op( OP_TERMINATE,0 );
  for( x=1 ; x &&
     (inp(at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|HOST_FLAG_3)) ; x++ )
    ; /* Wait for DSP to complete initialization */
  dsp_command( EXECUTE );
}

void show_image(void)
{
unsigned char buff[512];
unsigned max,min,page,row;
register int x,y;
long tick,z,t;
unsigned count;
int fd,s,u;

  while( !(inp(at200_base + DSP_INTERRUPT_STATUS) & HOST_FLAG_3) )
    ;   /* Wait till DSP is ready for IRQB */
  reset_fifo();
  dsp_command( DSP_IRQB );
  tick = dos_ticks() + 18L + (long) (exposure*time + shutter * 2 / 55); /* Timeout Delay */
  while( !(inp(at200_base + STATUS) & FIFO_EMPTY) )  /* Wait for 1st Pixel */
    if( dos_ticks() >= tick ) { /* Timeout! */
      mcga_display_mode( 0 );
      printf("Timeout without receiving a pixel!\n");
      exit( 1 );
    }
  ems_collect_image(ems_handle);    /* Use the fast routine for up to 64K Pixels */
  if( !(inp(at200_base + STATUS) & FIFO_OVERFLOW) ) /* Check for overflow */
    return;
  for( y=8192,min=65535,page=max=0,z=0 ; z< (262144/(p_bin * s_bin)) ; y++,z++ ) {
  	if(y >= 8192 ) {
  		ems_map_page(0,ems_handle,page++);
  		y=0;
  	}
      if( max < data[y] )
        max = data[y];
      if( min > data[y] )
        min = data[y];
    }
  if( max== min )
    max = 1;    /* Don't divide by zero */
    else
    max=0xFFFF / (max - min);

    t = ((512L*512L)/((long)p_bin * (long)s_bin));
    if(t > 64000L)
    t = 64000L;
    s = 512/s_bin;
    u = s;
    if(s > 320)
    s=320;
  for( y=8192,row=x=0,z=0,page=0 ; z < t ; x++,y++,z++ ) {
  	if(y >= 8192) {
  		ems_map_page(0,ems_handle,page++);
  		y -= 8192;
  	}
      buff[x] = (unsigned char) (((data[y] - min) * max) >> 8);
      if( x >= (s - 1) ) {
    memcpy(screen_row[row++],buff,s);
    x=-1;
    if( u > 320)
    y += (u - 320);
}
  }
/* write buffer to file picture */
      if(dofile) {
	fd = open(fname,O_WRONLY | O_CREAT |O_BINARY , S_IREAD | S_IWRITE);

  for( page=0 ; page<(unsigned)((32/(s_bin * p_bin))) ; page) {
  ems_map_page(0,ems_handle,page++);
for(x=0; x < (8*1024);x++)
  ptable[x] = data[x];
  x=0;
	count=write(fd,ptable,16*1024);
}
	close(fd);

}

}

void show_help(void)
{
register x;
char buff[81],number[8];

  printf("USAGE: display [-a #] [-b #] [-c #] [-e #] [-f ' '][-o #] [-s #][-t]\n"
         "  Where:\n"
         "    -a #  # is the base address descriptor (2 by default)\n"
         "    -b    indicates bin to see full field of view\n"
         "    -c #  # selects the CCD type from the following list:\n");
  strcpy(buff,"           ");
  for( x=0 ; x<CCD_COUNT ; x++ ) {
    sprintf(number," %u=",x);
    if( strlen(buff) + strlen(number) + strlen(ccd[x].id) < 80 ) {
      strcat(buff,number);
      strcat(buff,ccd[x].id);
    }
    else {
      printf("%s\n",buff);
      strcpy(buff,"           ");
      strcat(buff,number);
      strcat(buff,ccd[x].id);
    }
  }
  printf("%s\n",buff);
  printf("    -e #  # is the exposure time in milliseconds (>0)\n"
         "    -f `filename' is the name of the stored output image\n"
         "    -l is the number of times you wish to loop through the procedure\n"
         "    -o #  # is the open/close time for the shutter in milliseconds (>0)\n"
         "    -r #  # is the resolution of the A/D converter in bits (12,14,16)\n"
         "    -s #  # selects the Serial Control Byte where:\n"
         "            0=40KHz, 1=50KHz, 2=200KHz, 3=500KHz\n"
         "    -t    indicates you want to specify exposure in seconds\n");
  exit( 0 );
}

void main(int argc,char *argv[])
{
register unsigned x,y;
int data_bits = 12;

  for( x=1 ; x < (unsigned) argc ; x++ ) {
    if( argv[x][0] == '-' )
      switch( argv[x][1] ) {
        case 'a': case 'A':
          x++;
          at200_base = ((0x1F & atoi(argv[x])) << 4) | 0x200;
          break;
        case 'b': case 'B':
          x++;
          p_bin = atoi(argv[x]);
          s_bin = atoi(argv[x]);
          break;
        case 'c': case 'C':
          x++;
          ccd_type = atoi(argv[x]);
          if( ccd_type >= CCD_COUNT )
            show_help();
          break;
        case 'e': case 'E':
          x++;
          exposure = atoi(argv[x]);
          if( exposure <= 0 )
            show_help();
          break;
        case 'f': case 'F':
          x++;
          dofile=1;
          name=argv[x];
         break;
        case 'l': case 'L':
          x++;
          loop = atoi(argv[x]);
          if( loop <= 0 )
            show_help();
          break;
        case 'o': case 'O':
          x++;
          shutter = atoi(argv[x]);
          if( shutter <= 0 )
            show_help();
          break;
        case 'r': case 'R':
          x++;
          data_bits = atoi(argv[x]);
          if( data_bits != 12 && data_bits != 14 && data_bits != 16 )
            show_help();
          break;
        case 's': case 'S':
          x++;
          ser_byte = atoi(argv[x]);
          if( ser_byte < A40KHZ || ser_byte > A500KHZ )
            show_help();
          break;
        case 't': case 'T':
          time = 1000;
          break;
        default:
          show_help();
          break;
      }
    else
      show_help();
  }
  printf("Configuration:\n  %s CCD\n  ",ccd[ccd_type].id);
  switch( ser_byte ) {
    case A40KHZ:
      printf("40");
      x = load_dsp_program(dsp_cam40k,size_of_dsp_cam40k);
      ser_byte = 63;
      break;
    case A50KHZ:
      printf("50");
      x = load_dsp_program(dsp_cam50k,size_of_dsp_cam50k);
      ser_byte = 255;
      break;
    case A200KHZ:
      printf("%d bit 200",data_bits);
      if( data_bits == 16 )
        x = load_dsp_program(dsp_cam216,size_of_dsp_cam216);
      else
        x = load_dsp_program(dsp_cam200k,size_of_dsp_cam200k);
      ser_byte = 32;
      break;
    case A500KHZ:
      printf("500");
      x = load_dsp_program(dsp_cam500k,size_of_dsp_cam500k);
      ser_byte = 5;
      break;
  }
  printf("KHz Analog Card.\n");
  if( !x ) {
    printf("Error loading program.\n");
    exit( 1 );
  }
  for( x=1 ; x ; x++ )
    if( !(inp(at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|HOST_FLAG_3)) )
      break;  /* Wait for DSP to complete initialization */
  if( !x ) {  /* Timed out! */
    printf("The AT200 is unresponsive or missing.  Unable to proceed.\n");
    exit( 2 );
  }

ptable = (unsigned *) &buffer;
ems_handle = ems_alloc( 262144 * sizeof(unsigned) );
if( !ems_handle ) {
	printf("Unable to allocate EMS. Unable to procede.\n");
	exit(3);
}
data=(unsigned *) ems_get_page_frame(); /* Returns pointer to EMS page frame. */
if( !data ) {
	printf("Unable to locate EMS page frame. Unable to procede.\n");
	exit(4);
}
  for( x=0 ; x<200 ; x++ )    /* Load graphics display array */
    screen_row[x] = &((char far *)0xA0000000)[x * 320];
  load_ops();
  mcga_display_mode( 1 );
  for( x=0 ; x<64 ; x++ ) {
  	y = x*4;
    mcga_color_set(y,x,x,x);
    mcga_color_set(y+1,x,x,x);
    mcga_color_set(y+2,x,x,x);
    mcga_color_set(y+3,x,x,x);
}
  for( x=0 ; x < (unsigned)loop ; x++ ) {
    sprintf(fname, "%s%u",name,x);
    show_image();
    if( kbhit() ) {
      x = getch();
      break;
    }
}
  getch();
  mcga_display_mode( 0 );
  ems_free( ems_handle );
  exit( 0 );
}
